/* Heightfield Lab : generate and modify heightfields
 * intended to work in a manner similar to Lee Crocker's "piclab" program
 * Uses 32-bit floating point internally to store pixel values.
 * Saves to TGA, POT, PGM, PNG, GIF, and Matlab formats.
 * Copyright (C) 1996-2007  John P. Beale  <beale@best.com>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 * --------------   JPB 1/18/96  --------------------------------
 */

#include <stdio.h>                      /* fopen() sprintf() etc. */
#include <stdlib.h>                     /* strtod() */
#include <math.h>                       /* math stuff */
#include <string.h>                     /* strcpy() */
#include <ctype.h>                      /* tolower() */
#include <dirent.h>                     /* opendir, readdir */
#include <unistd.h>                     /* unlink() */
#include <time.h>                       /* ctime(), time_t */
#include "hl.h"
#include "hf_io.h"
/* #include "hl.h" */
#ifdef USE_X                            /* X Window graphics option */
#include "typy.h"
#endif

int strnicmp_jpb(char *a, char *b, int n);
int f_delete(char *fname);   /* delete a file on the disk */
int show_hist(int lines);
void list_vars();              /* list desc. of variables */
void show_vars();              /* print values of variables */
int h_tshow(int tx, int ty);   /* show matrix as text (ick) */
int h_newc(hfield *hfp, int xs, int ys, char *name); /* create complex HF */
int h_newr(hfield *hfp, int xs, int ys, char *name); /* create real HF */
void pr_usage(char *name);

/* ---------------------------------------------------- */
#define CMD_LINES 100
#define CMD_LEN   81

/* ----- global variables ---------------------------------- */

hfield *hst = NULL;     /* stack of heightfields */
int hp;                 /* pointer to top of stack: hst[hp] */
char cmd_hist[CMD_LINES][CMD_LEN];     /* command history buffer */    
int cmd_ptr = 0;                       /* pointer to cmnd history buffer */
char cmd_instr[CLINELEN];              /* the input command line */
int cmd_p;                             /* pointer to cmd_instr[] */

time_t time_start;      /* current time, stored by time_tic() */
int stacksize;          /* total size of stack */
int histbins;           /* histogram bins */
int tile_mode;          /* tile vs. non-tilable mode */
D   tile_tol;           /* tiling edge threshold tolerance */
int nskip;              /* 3-D viewing parameter: # of HF elements to skip */
int v_tile;             /* view 3D in tiled mode? */
int graph_mode;         /* graphics mode used for viewing (DOS) */
int xgsize,ygsize;      /* size of graphics viewing window (X) */
int shader;             /* 0 wireframe 1 height 2 slope 3 lightsource */
int sigbits;            /* number of bits to store in PNG file */
D sl_min;            /* minimum slope (shading offset) */
D sl_scale;             /* slope scale factor (shading scale) */
int rnd_seed_stale;     /* no random seed yet? */
int rnd_seed;           /* random seed to use (if not stale) */
int rand_gauss = TRUE;  /* if gen_rand() should use gaussian distribution */
int noxinit = TRUE;     /* if X Window has been created yet */
int logging = FALSE;    /* if command logging is turned on */
int takefile = FALSE;   /* if a command file is being read in */
int echo_cmd = TRUE;    /* if commands from take file should be echoed */
int startint = FALSE;   /* if HF-Lab session started in interactive mode */
int hist_on = TRUE;     /* if commands being recorded in buffer */
int int_done = FALSE;   /* interp-loop-exit flag */
int view_current = FALSE;   /* if currently in view mode */
int show_current = FALSE;   /* if currently in show mode */
PT3 light;              /* position of light source */
PT3 eye;                /* position of viewer */

FILE *logfp;            /* command logging file */
FILE *cmd_stream;       /* input stream (stdin or command file) */

/* --------------------------------------------------------- */
int main (int argc, char *argv[])
{
 char takefname[80] = "";
 int i;

 xgsize = 400; ygsize = 300;  /* initial X windows screen size */

  if (argc > 1) {  /* command line arguments are present */
   i = 1;
   do {  /* loop over all arguments */
    do { /* loop over all '-' options prior to a filename */
     if (argv[i][0]=='-') {    /*  an option of some kind */
      if (keymatch(argv[i],"-geometry",3)) {
	if (++i >= argc) { 
	    printf("-geometry option requires an argument.\n");
	    exit(1);
	}
	sscanf(argv[i++], "%dx%d", &xgsize, &ygsize);
	printf("(Xwindow set to %dx%d)\n",xgsize,ygsize);
      } else if (keymatch(argv[i],"-help",2)) {
	pr_usage(argv[0]);
	exit(0);
      } else {
	printf("HF-Lab: option '%s' not understood.\n",argv[i]);
	printf("Type '%s -help' for options.\n",argv[0]);
	exit(1);
      } /* end else */
     } /* end if (char=='-') */
    } while (i<argc && argv[i][0] == '-');

    if (i<argc) {
      if (!takefile) {   /* take filename already specified */
	takefile=TRUE;
	strcpy(takefname,argv[i]);
      } else {
	printf("Hf-Lab: %s: only one takefile allowed.\n",argv[i]);
	printf("Type '%s -help' for options.\n",argv[0]);
	exit(1);
      }
    } /* end if (i<argc) */
   } while (++i < argc);
  } /* end else (argc > 1) */

  if (takefile) {
    if ((cmd_stream = fopen(takefname,"r")) == NULL) {
      printf("HF-Lab error: unable to open cmd file %s\n",takefname);
      exit(1);
    } else {
    printf("HF-Lab: reading command file %s\n",takefname);
    startint = FALSE;
    }
  } else {      /* --- print welcome banner ---- */
   printf("HeightField Lab %s (C) 2007 John P. Beale\n",VERSION_NUMBER);
   printf("Type HELP for more information.\n\n");
   cmd_stream = stdin;  /* take input from keyboard */
   startint = TRUE;
  }
  /*  printf("initial arg = %s\n",argv[0]);  */

  /* --- overall initialization functions ------- */
  nskip = 1;         /* 3D viewing parameter */
  v_tile = FALSE;    /* tiled 3D display */
  graph_mode = 2;    /* graphics mode */
  shader = LIGHT;    /* shading */
  sl_min = 1.0;   /* slope shading offset */
  sl_scale = 1.0;    /* slope shading scale */
  sigbits = 16;      /* # bits to store in PNG greyscale file */
  tile_mode = AUTO;  /* tiling sensor */
  tile_tol = 0.01;   /* autodetect tiling tolerance value */
  histbins = 1000;   /* number of histogram bins */
  time_start = time(NULL); /* store program start time */
  rnd_seed = 0;
  rnd_seed_stale = TRUE; /* generate a new random seed */
  light.x = 1;
  light.y = 0;
  light.z = -1;
  eye.x = 0.0; eye.y = -2; eye.z = -3;
 
  if (clear_stack(4)) {
    printf("Could not initialize stack.\n");
    exit(1);
  }

  /* ------------------------------------main interp. loop ------- */
  do {
    
    do {                           /* get a non-null cmd line */
       if (!(takefile && !echo_cmd))
	printf("HL>");             /* input prompt for user entry */
       getline(cmd_instr,CLINELEN-1);  /* get one line of command(s) */
    } while (cmd_instr[0] == '\0');    /* now have non-null line */

    cmd_p = 0;    /* start parsing at the first character */
    cmd_seq();

  } while (!int_done);

  return(0);
}  /*--------  end of main() for HF-Lab ------ */

int clear_stack(int newsize)      /* erase all entries in HF stack */
{
 int i;
 size_t msize;

  if (hst != (hfield *)NULL) {  /* something on stack already */
    for (i=0;i<stacksize;i++) {    /* init all HFs in stack to null vals */
     if (hst[i].a != (PTYPE *)NULL) free(hst[i].a);
     hst[i].a = (PTYPE *)NULL;
    }
    free(hst);                  /* release old stack memory */
  }
  msize = sizeof(hfield) * (size_t) newsize;
  hst = (hfield *) malloc(msize); /* alloc stack array */

  if (hst == (hfield *)NULL) {
   fprintf(stderr,"Memory allocation for stacksize = %d failed.\n",newsize);
   return(1);
  }
  
  stacksize = newsize;
  for (i=0;i<stacksize;i++) {    /* init all HFs in stack to null vals */
     hst[i].a = (PTYPE *)NULL;
     hst[i].xsize = hst[i].ysize = 0;
     hst[i].min = 0;
     hst[i].max = 0;
     hst[i].c = FALSE;
     hst[i].name[0] = '\0';
  }
  
  hp = 0;               /* set hp to top of stack */
  return(0);
} /* clear_stack() */

void cleanup()  /* clean things up before exiting */
{
#ifdef USE_X
 if (!noxinit) close_graph();
#endif
 if (logging) fclose(logfp); logging=FALSE;
}

void show_vars()              /* print variable values */
{
  printf("histbins:    %d\n",histbins);
  printf("stacksize:   %d\n",stacksize);
  printf("nskip:       %d\n",nskip);
  printf("tolerance:   %1.3f\n",(float)tile_tol);
  printf("graphmode:   %d\n",graph_mode);
  printf("sl-min:      %1.3f\n",(float)sl_min);
  printf("sl-scale:    %1.3f\n",(float)sl_scale);
  printf("sigbits:     %d\n",sigbits);
  printf("seed:      %s %d\n",(rnd_seed_stale?" ":"*"),rnd_seed);
  printf("randgauss:   ");
  if (rand_gauss) printf("TRUE\n"); else printf("FALSE\n");
  printf("tile_mode:   ");
  if (tile_mode==AUTO) printf("AUTO\n");
  else if (tile_mode==TON) printf("ON\n");
  else printf("OFF\n");
  printf("v_tile:      ");
  if (v_tile) printf("TRUE\n"); else printf("FALSE\n");
  printf("echo_cmd:    ");
  if (echo_cmd) printf("TRUE\n"); else printf("FALSE\n");
  printf("shader:      ");
  if (shader==WIRE) printf("WIRE");
  else if (shader==HEIGHT) printf("HEIGHT");
  else if (shader==SLOPE)  printf("SLOPE");
  else if (shader==LIGHT)  printf("LIGHT");
  printf("\n");
}

int set_vars(char *var, char *arg)      /* set variable values */
{
 int t1;

  if (keymatch(var,"histbins",2)) {
    sscanf(arg,"%d",&histbins);
    if (histbins < 2) histbins = 2;
  } else if (keymatch(var,"stacksize",2)) {
    sscanf(arg,"%d",&t1);
    if (t1==stacksize) { 
      printf("Stacksize already %d. Stack unchanged.\n",t1);
      return(1);
    } else if (t1<2) {
      printf("Stacksize minimum is two.\n");
      return(1);
    } else if (t1 == 2) {
      printf("Warning: stack will be too small for some operations.\n");
    }
    if (clear_stack(t1)) 
      printf("Error: stack cleared. Stacksize %d.\n",stacksize);
      else printf("Stack cleared; stacksize set to %d\n",stacksize);
  } else if (keymatch(var,"nskip",2)) {
    sscanf(arg,"%d",&nskip);
    if (nskip < 1) nskip = 1;
  } else if (keymatch(var,"graphmode",2)) {
    sscanf(arg,"%d",&graph_mode);
    if (graph_mode < 0) graph_mode = 0;
    if (graph_mode > 3) graph_mode = 3;
  } else if (keymatch(var,"formats",2)) {
    pr_help("formats");
  } else if (keymatch(var,"randgauss",2)) {
    if (keymatch(arg,"TRUE",1)) rand_gauss = TRUE;
    else rand_gauss = FALSE;
  } else if (keymatch(var,"echo_cmd",4)) {
    if (keymatch(arg,"TRUE",1)) echo_cmd = TRUE;
    else echo_cmd = FALSE;
  } else if (keymatch(var,"v_tile",1)) {
    if (keymatch(arg,"TRUE",1)) v_tile = TRUE;
    else v_tile = FALSE;
  } else if (keymatch(var,"seed",2)) {
    sscanf(arg,"%d",&rnd_seed);
    rnd_seed = ABS(rnd_seed);
    rnd_seed_stale = FALSE;
  } else if (keymatch(var,"tolerance",2)) {
    sscanf(arg,"%lf",&tile_tol);
  } else if (keymatch(var,"sl-min",3)) {
    sscanf(arg,"%lf",&sl_min);
  } else if (keymatch(var,"sl-scale",3)) {
    sscanf(arg,"%lf",&sl_scale);
  } else if (keymatch(var,"sigbits",3)) {
    sscanf(arg,"%d",&t1);
    if ((t1<1) || (t1>16)) 
      printf("Error: sigbits must be between 1 and 16 inclusive.\n");
    else sigbits = t1;
  } else if (keymatch(var,"tile_mode",3)) {
    if (keymatch(arg,"AUTO",1)) tile_mode = AUTO;
    else if (keymatch(arg,"ON",2)) tile_mode = TON;
    else if (keymatch(arg,"OFF",2)) tile_mode = TOFF;
    else printf("possible tile_mode values: ON OFF AUTO\n");
  } else if (keymatch(var,"shader",3)) {
    if (keymatch(arg,"WIRE",1)) shader = WIRE;
    else if (keymatch(arg,"HEIGHT",1)) shader = HEIGHT;
    else if (keymatch(arg,"SLOPE",1)) shader = SLOPE;
    else if (keymatch(arg,"LIGHT",1)) shader = LIGHT;
    else printf("possible shader values: WIRE HEIGHT SLOPE LIGHT\n");
  } else {
     printf("Variable name not found. These are the variables:\n");
     list_vars();
   }
  return(0);
}

int pr_help(char *cmd)              /* print help message */
{ char *hfile;
  char line[120];
  char ss[80];     /* command search string */
  int len;         /* length of search string */
  int i;           /* utility counter var */
  FILE *fp;

   hfile = getenv("HFLHELP");
   if (hfile == NULL) hfile = "hf-lab.hlp";
   fp = fopen(hfile,"r");
   if (fp == NULL) {
    printf("Unable to open helpfile %s. Check that you have set\n",hfile);
    printf("the environment variable HFLHELP = .../hf-lab.hlp\n");
    return(1);
   }
   if (cmd[0]=='\0') {  /* no command: use generic help message */
    do {                
       if (fgets(line,80,fp) == NULL) break;
       if (line[0] == '#') break;       /* marker for end of generic help */
       len = strlen(line);
       line[len-1]='\0';  /* change the \n char to a null */
       printf("%s",line);
       if (line[0]=='\t') len += 7; /* count extra spaces from tab */
       if (line[1]=='\t') len += 7; /* count extra spaces from tab */
       for (i=0;i<(79-len);i++) printf(" "); /* fill blank space to EOL */
       printf("\n");
    } while (TRUE);
   } else {             /* print specific help message */
     ss[0] = '#';
     strcpy(&ss[1],cmd);
     len = strlen(ss);
     printf("\n");      /* leading blank line */
     do {
	if (fgets(line,80,fp) == NULL) { /* reached end of file */
	  printf("Sorry, no help available for '%s'\n\n",cmd);
	  return(1);
	}
	if (strnicmp_jpb(line,ss,len) == 0) {  /* found it! */
	  do {
	    if (fgets(line,80,fp) == NULL) {
		printf("-- Ahem. Helpfile appears to be corrupted.\n");
		return(1); }
	  } while (line[0] == '#'); /* skip over synonym listings */          
	  do { /* print out help message */
	       printf("%s",line);
	       if (fgets(line,80,fp) == NULL) {
		  printf("-- Ahem. Helpfile appears to be corrupted.\n");
		  return(1);
	       }
	       if (line[0] == '#') return(0); /* end */
	     } while (TRUE);    /* printing this help */
	} /* if found */
     } while (TRUE);     /* search help file */
   }  /* getting specific help msg. */
   fclose(fp);
  return(0);
} /* pr_help */

int show_hist(int lines)
{
 int i, tp, lout;

 if (lines >= CMD_LINES) {
   tp = (cmd_ptr+1) % CMD_LINES;  /* show all of stack */
   lines = CMD_LINES-1;
   printf("---Command buffer is %d lines:\n",lines);
 } else {
   tp = ( (cmd_ptr - lines) + CMD_LINES ) % CMD_LINES;
   printf("  ---command history:\n");
 }

 lout = 0;
 for (i=0;i<lines;i++) {
   if (cmd_hist[tp][0] != '\0') {
     printf("  %s\n",cmd_hist[tp]);
     if (++lout == 23) {
       printf("----------<return> for more-------");
       getc(stdin);
       lout = 0;
     }
   }
   tp = (tp+1) % CMD_LINES;
 }
 return(0);
} /* show_hist */

int getline(char s[], int lim)       /* read in one line */
{ int c, i;
  c = 0;

  for (i=0; i<lim-1 && (c=getc(cmd_stream))!=EOF && c!='\n'; ++i)
     s[i] = c;
  s[i] = '\0';
  if ((i==0) && (c==EOF) && (takefile)) {
   strcpy(s,"quit");            /* turn EOF into "quit" */
   i = strlen(s);
  }
  if (hist_on && (s[0]!='\0')) {  /* cmd history: don't record blank lines */
    if (++cmd_ptr >= CMD_LINES) cmd_ptr = 0;
    strncpy(cmd_hist[cmd_ptr],s,CMD_LEN-1);
    cmd_hist[cmd_ptr][CMD_LEN-1] = '\0';   /* insure null termination */
  }

#ifdef USE_DOS
  if (takefile && kbhit()) {
    printf("Script interrupted; wait...\n");
    fclose(cmd_stream); 
    cmd_stream=stdin;
    takefile = FALSE;
  }
#endif

  if (takefile && echo_cmd) printf("%s\n",s); /* show command from file */
  if (logging) fprintf(logfp,"%s\n",s);  /* record in file */
  return(i);
}

/* Case-insensitive keyword matcher from ppm package.
 * returns 1 if first minchars of str are identical to keyword
 * (ignoring case). Returns 0 if this is not the case.
 */

int keymatch( char *str, char *keyword, int minchars )
{
    int len, len2;
    char c1, c2;

    /* printf("matching %s && %s\n",str,keyword);     */
    len = strlen( str );              
    len2 = strlen( keyword);
    if (len2 < len) len = len2;    /* len is lesser of two string lengths */
    if ( len < minchars ) return 0;
    while ( --len >= 0 )
    {
	c1 = *str++;
	c2 = *keyword++;
	if ( c2 == '\0' ) return(0);              /* ran out of characters */
	if (isupper( c1 )) c1 = tolower( c1 );   /* force to lowercase */
	if (isupper( c2 )) c2 = tolower( c2 );
	if ( c1 != c2 ) return(0);                /* fail at first non-match */
    }
    return(1);
} /* keymatch() */

int s_empty()   /* returns true if hp points to empty slot */
{
  if (hst[hp].a==(PTYPE *)NULL) {
    printf("Oops: stack empty.\n");
    return(1);
  }
  else return(0);
}

int loadhf(char *fname_in)
{
 int rval;
 hfield h0;
  
   rval = readn_hf(&h0, fname_in);  /* readn_hf mallocs memory */
   if (rval != 0)  {
     printf("Unable to load file %s.\n",fname_in);   
     return(1);
   }
   strncpy(h0.name,fname_in,14);
   h0.name[14] = '\0';                  /* insure null termination */
   
   INCSTACK;
   if (hst[hp].a != (PTYPE *)NULL) free(hst[hp].a);  /* erase old hf */

   hst[hp] = h0;
   h_minmax();
   printf("read in %s: [%d x %d] ",fname_in,h0.xsize,h0.ysize);
   printf("range %1.2f .. %1.2f",hst[hp].min,hst[hp].max);
   if (h0.c) printf(" (complex)");
   printf("\n");
   return(0);
} /* loadhf() */

int savehf(char *fname_out)
{ 
 int rval, i, pos;
 char ext[5];
  
  if (s_empty()) return(1);     /* check if stack is empty */
   if ((hst[hp].min<0)||(hst[hp].max>1)) {
     pos = strlen(fname_out)-4;                /* check extension */
     if (pos < 1) {
       printf("File type not recognized.\n");
       return(1);
     }
     for (i=0;i<4;i++) ext[i]=fname_out[pos+i];
     ext[4]='\0';
     if ( !s_equal(ext,".MAT") && !s_equal(ext,".OCT")) {
       printf("Error: out of range. Use NORM to bring data into 0..1 range\n");
       printf("or save in MAT or OCT formats to preserve full range.\n");
       return(1);
     }
   }
   printf("Writing %s: [%d x %d]\n",fname_out,hst[hp].xsize,hst[hp].ysize);
   rval = writen_hf(hst[hp], fname_out);
   return (rval);
} /* savehf() */

int startlog(char *logname)
{
 time_t now;

  if (takefile) printf("Warning: Logging while taking input from file.\n");
  if (logging) {
	printf("Logging is already turned on.\n");
	return(1);
  }
  if ((logfp = fopen(logname,"a"))==NULL) {   /* open file for append */
    printf("Unable to open log file: %s\n",logname);
    return(1);
  }
  now = time(NULL);
  fprintf(logfp,"# HF-Lab %s Logfile started %s",VERSION_NUMBER,ctime(&now));
  logging = TRUE;
  printf("Logging to file %s\n",logname);
 return(0);
} /* startlog() */

int stoplog()
{
 if (!logging) {
	 printf("Logging has not been turned on.\n");
	 return(1);
 }
 fprintf(logfp,"# ------\n");
 fclose(logfp);
 logging = FALSE;
 printf("Logfile closed.\n");
 return(0);
} /* stoplog() */

 /* compare n chars, ignore case. 1 if unequal, 0 if equal. */
int strnicmp_jpb(char *a, char *b, int n) 
{
 int i;

 n = MIN(strlen(a),n);  /* examine at most n chars */
 n = MIN(strlen(b),n);
 for (i=0; i<n; i++) {
  if (tolower(a[i])!=tolower(b[i])) return(1);
 }
 return(0);
} /* strnicmp_jpb */

 /* compare strings, ignore case. 1 if equal, 0 if unequal. */
int s_equal(char *a, char *b)
{
 int i,n;

 n = MIN(strlen(a),strlen(b));
 for (i=0; i<n; i++) {
  if (tolower(a[i])!=tolower(b[i])) return(FALSE);
 }
 return(TRUE);
} /* s_equal */


#define P(string) printf("%s\n",string);

void list_vars()
{
  P("------------HF-Lab internal variables: 'set <varname>'---")
  P("")
  P("nskip:      matrix elements to skip in 3D viewing mode")
  P("v_tile      TRUE: view tiled HF   FALSE: view single HF")
  P("graphmode:  DOS graphics mode [0:320x200 1:640x480 2:800x600 3:1024x768]")
  P("shader:     view 3D shading model [WIRE HEIGHT SLOPE LIGHT]")
  P("sl-min      slope to set to black (eg, minimum slope) in SLOPE model")
  P("sl-scale    range of HF slopes to span the colors black...white")
  P("sigbits     number of bits to store in PNG or POT format files (1..16)")
  P("seed        rand# seed used last, or if '*' present, to be used next")
  P("")
  P("tile_mode:  treatment of matrix edge-wraparound [ON OFF AUTO]")
  P("stacksize:  number of HFs to store on stack. Changing clears stack.")
  P("tolerance:  threshold for HF to be tilable (if tile_mode = AUTO)")
  P("echo_cmd:   if commands from a script file should be echoed on screen")
  P("randgauss:  RAND distribution function: TRUE: normal  FALSE: uniform")
  P("histbins:   the number of bins used by EQUALIZE (histogram equalization)")
  P("")
}

int h_tshow(int txsize, int tysize)  /* show matrix as text (ick) */
{
int ix,iy,xx,yy;
int xsize, ysize;
int pels;             /* number of pixels averaged */
D scalefac;
D hmin, hmax, tmp;
D xblock,  yblock;
PTYPE *hf;
char shade[11] = " .-+*o%#@H";     /* text 'greyscale' levels */

  if (s_empty()) return(1);     /* check if stack is empty */
  hf = hst[hp].a;
  xsize = hst[hp].xsize;
  ysize = hst[hp].ysize;
  hmin = hst[hp].min;
  hmax = hst[hp].max;
  if (hmax == hmin) {
   printf("Error: tshow: cannot show constant matrix (min = max).\n");
   return(1);
  }
  scalefac = 9.999 / (hmax - hmin);
  xblock = (D)xsize/txsize;
  yblock = (D)ysize/tysize;
  for (iy=0;iy<tysize;iy++) {
   for (ix=0;ix<txsize;ix++) {
    tmp = 0;
    pels = 0;
    for (xx=0;xx<xblock;xx++) {
      for (yy=0;yy<yblock;yy++) {
	tmp += El(hf,(int)(ix*xblock+xx),(int)(iy*yblock+yy));
	pels++;
      }
    }
    tmp = ((tmp/pels) -hmin)* scalefac;
    printf("%c",shade[(int)tmp]);
   } /* for ix */
   printf("\n");
  } /* for iy */

  return(0);
} /* tshow() */

int h_newr(hfield *hf, int xs, int ys, char *name) /* create real HF */
{
 PTYPE *a;
 size_t mem;
 hfield h1;
 
 h1 = *hf;
 mem = (size_t)xs*ys*sizeof(PTYPE);
 a = (PTYPE *)malloc(mem);
 if (a == (PTYPE *)NULL) {
   printf("Error: out of memory.\n");
   return(1);
 }
 memset(a,0,mem);       /* initialize HF to all zeros */
 h1.a = a;
 h1.xsize = xs;
 h1.ysize = ys;
 h1.c = 0;
 strncpy(h1.name,name,14);
 h1.name[14] = '\0';                  /* insure null termination */
 h1.max = h1.min = 0;

 *hf = h1;
 return(0);
} /* h_newr() */

int h_newc(hfield *hf, int xs, int ys, char *name) /* create real HF */
{
 PTYPE *a;
 size_t mem;
 hfield h1;
 
 h1 = *hf;
 
 mem = (size_t)2*xs*ys*sizeof(PTYPE);
 a = (PTYPE *)malloc(mem);
 if (a == (PTYPE *)NULL) {
   printf("Error: out of memory.\n");
   return(1);
 }
 memset(a,0,mem);       /* initialize HF to all zeros */
 h1.a = a;
 h1.xsize = xs;
 h1.ysize = ys;
 h1.c = 1;
 strncpy(h1.name,name,14);
 h1.name[14] = '\0';                  /* insure null termination */
 h1.max = h1.min = 0;
 *hf = h1;
 return(0);
} /* h_newc() */

void pr_usage(char *name)
{
 printf("HF-Lab %s\n",VERSION_NUMBER);
 printf("usage: %s [options] <script_file>\n",name);
 printf("       options: -geometry <xsize>x<ysize>  (X version only)\n");
 printf("\n");
 printf("Type 'help' from within HF-Lab for more help.\n");
}
